/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

 
//                          ett version: 1.0


//  xxxxxxxxxxxxxxx    global variable declarations  xxxxxxxxxxxxxxxx

 
// These variables below can be directly modified by the user on the options form: 

var item_cnt= 5;       // number of stimuli in the task
  
var timer_enabled = false;   // if true => timer is enabled. if false timer is disabled. Exactly: the timer is always on

var timer = 600;     // Task time in seconds if clock = 1000ms (default). 

var is_back = true;    // if (is_back==true) => user can go back to previous task/item.  

var prg_type = '0';   //  type of information in Progress Bar ('0' => Items, '1' => Time; '2' => Nothing) 
    		
var demo = false;     // if (demo == true) => Show task solutions 

var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 

// Other global constants and variables:

const clock = 1000;  // basic timer resolution in ms

var timer_id = null;   // timer_id is the identifier of the repeated action you want to cancel. This ID is returned from setInterval().

var counter = 0;      // timer counter. if (counter >= time) Timer will call theEnd

var is_task = false;    // task in progress or not for pageResize() event. is_task == true for all events between startTask() and theEnd() events

var dirty = false;  // if  (dirty==true) => user changed any options in form 

var item = 0;   // current item in task     

var is_storage= false; // local storage is enabled on user browser or not

var is_move= false;  // user can move (drag) slider or not (see onMouseDown, onMouseUp, mouseMove events)

//  global arrays declarations..
 
/* stimulus => contains polar data for each item in range 0..item_cnt-1 
	 for each item in range 1.dim => 0..item_cnt-1; 2. dim => 0 == x0, 1 == y0, 2== polar radius, 3 = angle (radians) 
	 (0..2 units are percents, 3 = radians) */
var stim = [];    

/* for each item contains x,y pixels coordinates for 3 circles. 
   1. dim = item; 2 dim => circle number 0..2;  3.dim => 0==x, 1==y 
   xy array dimensions => [item] x [circles] x [x-coord,y-coord] => [0..item_cnt-1] x [0..2] x [X pixels, Y pixels] */   
var xy= [];

// dependent variable (error) for each item in range 0..item_cnt-1. 
var dv = [];

// missing values for each item in range 0..item_cnt-1
var mis = [];

/* for each item: last position of blue circle (units are percents of canvas.width + canvas.height). 
   used in drawStim() for scale repainting if pageResize() event is raised */
var blue= [];  
 

/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */


// application init [create/start Timer, check storage, show start panel..] 
function init() 
{    
  if (timer_id)  {   
    window.clearInterval(timer_id);  
  }  
  timer_id= setInterval(function () {appTimer()},clock);    
  is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }				  
	setIns();		
	showStart();     
}


// application timer handler function
function appTimer()  
{    
	if (is_task && timer_enabled) // else: nop, idle, do nothing 
	{
		counter++;  
		if (counter >= timer)  {     // task time is limited
			theEnd();
			return;
		}	
		if (prg_type == '1')   {   // '1' => Time. display time on Progress Bar  
			let rest= timer-counter;   // this is remaining time 
			let percent= Math.round(100*rest/timer);    
			document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
			document.getElementById('nav_items').textContent= stringTime(rest);    
		}
	}  
}	


// return integer into string hh:mm:ss time format
// input par: total == integer (time in seconds)
function stringTime(total) 
{
  var hh = Math.floor(total/(60*60));         
  var div_min = total % (60 * 60);
  var mm = Math.floor(div_min / 60);       
  var div_sec = div_min % 60;
  var ss = Math.ceil(div_sec);    
  // These lines ensure you always have two-digits. This formats your string to HH:MM:SS
  if (hh < 10)  {hh = "0"+hh;}     
  if (mm < 10) {mm = "0"+mm;}
  if (ss < 10) {ss = "0"+ss;} 
  return hh + ":" + mm + ":" + ss;
} 


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// raised when page is resized. see html: <body onresize="pageResize()"
// set task layout
function pageResize()  
{ 
	if (is_task)  { 
		let canvas= document.getElementById('canvas');
		canvas.width= window.innerWidth;  
		canvas.style.width= canvas.width + 'px';	     
		canvas.height= window.innerHeight - document.getElementById('footer').offsetHeight; // document.getElementById('footer').clientHeight == document.getElementById('footer').offsetHeight  
		canvas.style.height= canvas.height + 'px';  // https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/ 		
		
		drawStim();
	}
}	


// show start panel 
function showStart()  
{      
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('hide_option').style= 'text-align:center';  // bug? auto-remove style when element.style.display="none"
  document.getElementById('instructions').style.display= 'none';     
  if (is_storage && localStorage.getItem('ett-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('ett-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' ETT results stored in your web browser.';
  }
  dirty= false;
  //document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').classList.add('center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';          
  document.getElementById('hide_option').style.display = 'none';   
  //document.getElementById('start_div').style= 'text-align:center';  // bug?  (firefox auto remove class "center" when set style.display.none;)
  document.getElementById('show_option').style= 'text-align:center';  
  document.getElementById('instructions').style.display= 'block';
  refMsg(); 
  setIns();	
}


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		       
  document.getElementById('ins_demo').style.display= demo ? 'block' : 'none';	
	document.getElementById('ins_back').style.display= is_back ? 'block' : 'none';
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty=true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{
		
		case 'item_cnt':
			if (e.validity.valid)  {
				msg.textContent= 'The number of stimuli in the task'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break;  				
						
    case 'timer_enabled':   
      if (e.checked)  {
        msg.textContent= 'Yes';
        document.getElementById('p_timer').style.display= 'block';
      }
      else  {  
        msg.textContent= 'Timer is disabled';
        document.getElementById('p_timer').style.display= 'none'; 
      }       
      break;
		
    case 'timer':
			if (e.validity.valid)  {
				msg.textContent= stringTime(e.value); 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break;  		
			   		
    case 'is_back':     		
			msg.textContent= e.checked ? 'Yes' : 'No. Returning to previous items is not allowed.';    
			break;
        
    case 'prg_type':                                                                                                                                            
      if (e.value == '0')  {
        msg.textContent= 'It will be displayed how many items are left to the end.';
      }
      else if (e.value == '1')  {
        msg.textContent= 'The remaining time until the end of the task will be displayed. Only when timer is enabled.';
      }
      else  {
        msg.textContent= ''; //'Nothing';
      }
      break;    
							
		case 'demo':
       msg.textContent= e.checked ? 'On. The lengths of the sides of the triangle will be displayed. The percentage error will also be displayed.' : 'Off';  
       break; 
		
		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty === true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity.valid object in all input type="number"
function optionsValid()  
{ 
  let res= document.getElementById('item_cnt').validity.valid && document.getElementById('timer').validity.valid; 	
  return res;              
}


// start task (research) 
function startTask(e)   
{       			
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';              
	document.getElementById('task_section').style.display= 'block'; 	  
	document.getElementById('canvas').style.display= 'block';			    
		
	document.getElementById('progress').style.display= 'block';   //document.getElementById('progress').setAttribute('style', 'width:100%');	 
	if ( document.getElementById('prg_type').value == '2' || ( document.getElementById('prg_type').value == '1' && timer_enabled == false ) ) {
		document.getElementById('progress_holder').style.display= 'none'; 
	}	
	else  
	{
		document.getElementById('progress_holder').style.display= 'block';     
		document.getElementById('progress').setAttribute('style', 'width:100%');   		
		if (document.getElementById('prg_type').value == '0') {
			document.getElementById('progress').className='progress-bar progress-bar-info progress-bar-striped';     // items/total
		}
		else if (document.getElementById('prg_type').value == '1')  {
			document.getElementById('progress').className= 'progress-bar progress-bar-danger progress-bar-striped';  // remaining time   
		} 	
		document.getElementById('progress').setAttribute('style', 'width:0');  // reset progress bar width
	}					
	document.getElementById('nav_items').textContent='';  // reset   	
 
  setPar();
 
	stimGen();
			
  is_task= true;    // finally "activate" timer  	
  	
  pageResize();  // set default task layout
	
	drawStim();    // show first item (stimulus)
	  	 
}

  
// set internal application research parameters (global vars)
function setPar()  
{  
	item= 0;  			   
	counter= 0;       // reset timer counter
	
	stim= [];	 
	xy= []; 	
	dv= [];
	mis= []; 
  blue= [];	
  for (let i= 0; i< item_cnt; i++)   
  { // i => item
		stim[i]= [];
		xy[i]= [];	
    mis[i]= true; 
    blue[i]= [];		
		for (let j = 0; j < 3; j++)  {     
      xy[i][j]= [];			
    }		
  } 
}


// https://varun.ca/polar-coords/
// generate all stimuli for each item in range 0..item_cnt-1
function stimGen()	
{			 	
	let x0, y0, r, ang, prior=[-1,-1];		
	for (let i = 0; i < item_cnt; i++)    // i => item  
	{		
		do   {				
			x0 = getRndInt(40,60);  // offset from canvas horizontal center in x direction in percents is +/- 10%
			y0 = getRndInt(40,60);  // offset from canvas vertical center in y direction in percents is +/- 10%
		}  while ( Math.abs (x0 - prior[0]) < 5  &&  Math.abs (y0 - prior[1]) < 5 );  // 5% is min offset from previous item in x, y axis	
		prior= [x0, y0];
		stim[i][0]= x0; // units: percents canvas.width		  
		stim[i][1]= y0;	// units: percents canvas.height
		r = Math.min( 50 - Math.abs(50 - x0) , 50 - Math.abs(50 - y0) ); // line polar radius in percent of min (width,height)
	  r = getRndFloat(0.75 * r, r);  // you can comment this line if you want to have const max polar radius 
		ang= getRndFloat(0, 2*Math.PI); // 1. point => angle in radians
		stim[i][2]= r;  // polar radius in units: percents available min(width,height) of canvas 		  
		stim[i][3]= ang;   // polar angle in radians		
		
		dv[i]= demo ?  0 : null;		
	}	
}


// get random integer from min to max range
function getRndInt(min, max) {
	return Math.floor(Math.random() * (max - min + 1)) + min;
}


/* https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Math/random
  Returns a float random number between the specified values. 
  The returned value is no lower than (and may possibly equal) min, and is less than (and not equal) max.
  Returns a random number between min (inclusive) and max (exclusive)  */
function getRndFloat(min, max) {
  return Math.random() * (max - min) + min;
}


// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   
	
	if (document.getElementById('item_cnt').validity.valid)  {  
	  item_cnt= parseInt(document.getElementById('item_cnt').value, 10);			
  } 
		
	timer_enabled= (document.getElementById('timer_enabled').checked) ? true: false;  // timer is enabled or disabled
	
	if (timer_enabled && document.getElementById('timer').validity.valid) {
		timer= Math.round(parseInt(document.getElementById('timer').value, 10));  // task time (s)
  }	
	
	is_back= document.getElementById('is_back').checked ? true: false;   // enable Go back (previous item)      
  	
	prg_type= document.getElementById('prg_type').value;   //  type of information displayed on Progress Bar	
		
  demo= document.getElementById('demo').checked ? true: false;	
	
  store= (document.getElementById('store_group').disabled === false && document.getElementById('store').value === '1') ? true : false;

}


// goto previous item (stimulus)
function priorItem()  
{      
  if (item > 0)  {
    item--;  
    drawStim();     // draw again previous stimulus
  }         
}


// goto next item (stimulus)
function nextItem()  
{    		
	item++;					
	if (item < item_cnt)  {         		 
    drawStim();     // draw next stimulus => range 0..item_cnt-1        
  }
  else  {               
    theEnd();  // end of task if item >= item_cnt 
  }
}


// draw (repaint) current stimulus           
function drawStim() 
{    		
	
	let canvas= document.getElementById('canvas'), ctx= canvas.getContext('2d'), w= canvas.width, h= canvas.height;	
	ctx.clearRect(0, 0, w, h);  //ctx.clearRect(0, 0, canvas.clientWidth, canvas.clientHeight);
  
	ctx.strokeStyle= '#000';  //ctx.fillStyle= '#000';
	
	let x, y, points= [];		   	  	 	  	
	let x0 = stim[item][0] * w / 100;  // transform units: percent of canvas.width => pixels
	let y0 = stim[item][1] * h / 100;	 // transform units: percent of canvas.height => pixels				
	let r = stim[item][2] * Math.min(w,h) / 100 - 10;  // transform polar radius from percent => pixels
	let	ang = stim[item][3];   // polar angle (1. point) in radians 
	for (let j= 0; j<=2; j++)   
	{
		if (j > 0) { 
			ang+= 2*Math.PI/3; // increase angle in radians (120 degrees)
		}  			
		if (j == 2 && !demo) {
			ang= Math.random() < 0.5 ? ang + 2*Math.PI/6 : ang -  2*Math.PI/6; // default error in each task= +/- 2*Math.PI/6
		}
		ctx.beginPath();
		if ( j == 2 && mis[item] == false /*&& xy[item][2].length == 2*/ )  {
			// convert percent of canvas.width (canvas.height) into x (y) pixels 			
			x= blue[item][0] * w / 100;	 // convert percent of canvas.width into x pixels
      y= blue[item][1] * h / 100;  // convert percent of canvas.height into y pixels
		}
		else  {  // show default layout
	    x = x0 + r * Math.cos(ang);    
			y = y0 + r * Math.sin(ang);				
		}	
		xy[item][j]= [x,y];  
		ctx.arc(x, y, 10, 0, 2*Math.PI, false);  //ctx.arc(x,y,r,sAngle,eAngle,counterclockwise);            
		ctx.fillStyle= j == 2 ? '#00f' : '#000';
		ctx.fill();
	}
	
	ctx.fillStyle= '#000';  
	ctx.lineWidth= demo ? 1 : 3; 
	ctx.beginPath();  	 
	ctx.moveTo(xy[item][0][0], xy[item][0][1]);   // 0. point  
	ctx.lineTo(xy[item][1][0], xy[item][1][1]);   // draw line between 0. point and 1. point 
	if (mis[item]) {
		ctx.lineTo(xy[item][2][0], xy[item][2][1]);   // draw line between 1. point and 2. point     
	}
	else  {
		ctx.lineTo(blue[item][0] * w / 100, blue[item][1] * h / 100);   // draw line between 1. point and 2. point    
	}
	// ctx.closePath();
	ctx.lineTo(xy[item][0][0], xy[item][0][1]);   // draw line between 2. point and 0. point
	ctx.stroke();
	
	/* debug start		
	ctx.beginPath();
	ctx.moveTo(x0-3,y0);  
	ctx.lineTo(x0+3,y0);   // draw line on screen canvas 
	ctx.moveTo(x0,y0-3);  
	ctx.lineTo(x0,y0+3);   // draw line on screen canvas 
	ctx.stroke();
	
	ctx.beginPath();
	ctx.arc(x0, y0, r, 0, 2*Math.PI, false);  //ctx.arc(x,y,r,sAngle,eAngle,counterclockwise);            
	ctx.stroke();
	  debug end  */	
    
	let d01= getDistance(xy[item][0][0], xy[item][0][1], xy[item][1][0], xy[item][1][1]);   
	let d02= getDistance(xy[item][0][0], xy[item][0][1], xy[item][2][0], xy[item][2][1]);
	let d12= getDistance(xy[item][1][0], xy[item][1][1], xy[item][2][0], xy[item][2][1]);
	let opt= 3 * d01; // optimal triangle perimeter = ( 3 * a )

	dv[item]= 100 * Math.abs ( d01 + d02 + d12 - opt ) / opt;  // unit error percents 
	
	if (demo)  {		
		ctx.fillStyle= '#f00';  // font color 
  	ctx.font = '1em Arial';  // 'Georgia, serif'
		ctx.textAlign = 'center';
		ctx.textBaseline= 'bottom';
		x = divLine(xy[item][0][0], xy[item][1][0], -1);   
		y = divLine(xy[item][0][1], xy[item][1][1], -1); 
		ctx.fillText(d01.toFixed(2), x, y);  // strokeText actually strokes the outline of the letters while fillText fills the inside of the letters.
				
		x = divLine(xy[item][0][0], xy[item][2][0], -1);   
		y = divLine(xy[item][0][1], xy[item][2][1], -1); 
		ctx.fillText(d02.toFixed(2) + ' Error= ' + dv[item].toFixed(2) + '%', x, y);
				
		x = divLine(xy[item][1][0], xy[item][2][0], -1);   
		y = divLine(xy[item][1][1], xy[item][2][1], -1); 
		ctx.fillText(d12.toFixed(2), x, y);
  	
	} 		
	
  updateFooter();	  
}


// return distance from point P1(x1,y1) to point P2(x2,y2)
function getDistance(x1, y1, x2, y2)  {
  return Math.sqrt( Math.pow(x1-x2, 2) + Math.pow(y1-y2, 2) ); // total line distance (px) 
}	


/* divide line 
  first => first point (P1) 
  second => second point (P2)
  n => (distance(P1-Q)/ distance(Q-P2)) = lambda = -m/n = negative ratio length: P1-Q/Q-P2
  returns: coordinate of point (Q) between first and second point   */   
function divLine(first,second,n)  {     
  return (first-n*second)/(1-n);   // Math.round((first-n*second)/(1-n));
}


// update footer when show each stimulus in task. update buttons and update progress bar
function updateFooter()
{  
	document.getElementById('prior').style.display= item > 0 && is_back ? 'inline': 'none';  // set footer navigation
	let btn= document.getElementById('next');     
	if (item == item_cnt-1)  {       
		btn.style.backgroundImage=  "url('img/end.png')";  // update 'Next item' image on 'right arrow' button 
		btn.title= 'End of task';
	}
	else  {
		btn.style.backgroundImage= "url('img/forward.png')"; 
		btn.title= 'Go forward';
	}						
  if (prg_type == '0' )  {    // update Progress bar => show current item count and remaining item count on Progress Bar (default)                                                                                                                                                       
		let percent = Math.round(100*(item_cnt-(item+1))/item_cnt);      
		document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
		document.getElementById('nav_items').textContent= 'Item= ' + (item + 1) + ' Remaining= ' + (item_cnt - item-1);  
	}
}


// end of task (research)
function theEnd()  
{   

  is_task= false;    	
	
	document.getElementById('task_section').style.display= 'none';	
	document.getElementById('canvas').style.display= 'none'; 
	//document.getElementById('footer').style.display= 'none';
  document.getElementById('end').style.display= 'block';   
  document.getElementById('result_wrapper').style.display= 'block'; 
	
  let blank= 0, rows = [], err, arr= [];  
	// compute points (and total score) for each item in task.  
	for (let i = 0; i < item_cnt; i++)  
	{		
		if ( dv[i] == null )  {
      err= '';
		}
		else {
			err= dv[i].toFixed(2); 
			arr.push(dv[i]);
		}
		if (mis[i] == true)  {
      rows[i]= [ parseInt(i+1,10), err, 'No' ];  // => [ 'Item', 'Error (%)' , 'Moved' ]  
			blank++;			
		}
		else {	
			rows[i]= [ parseInt(i+1,10), err, 'Yes' ];  // => [ 'Item', 'Error (%)' , 'Moved' ]      
		}		
	}		

	createTableArray(document.getElementById('table1'), rows,  ['Item', 'Error (%)', 'Moved']); 
	 		
	document.getElementById('table1').style.display= 'block';
		
	// calculate decriptive statistics for the errors (%) => table2 + chart1 
	rows = [];
	let n= arr.length
	let m = (n > 0) ? mean(arr).toFixed(2) : '';     // mean only for valid items in task
	let sd = (n > 1) ? stDev(arr, m).toFixed(2): ''; // standard deviation only for valid items in task  	
	
	let tmp= ['Variable', 'Average', 'Sample Standard Deviation', 'N']; // header
	rows.push( ['Error (%)', m, sd, n] );  // last row
	
	createTableArray(document.getElementById('table2'), rows, tmp);   
	
  // "chart" => repaint "chart" 
  let ok= (100-m).toFixed(2);	
	document.getElementById('chart_2').setAttribute('style', 'width:' + m + '%');
	document.getElementById('chart_1').setAttribute('style', 'width:' + ok + '%');    // !important
	document.getElementById('chart').title= 'The average error is= ' + m + '%';  	
		
	// update 'score' label => above "Repeat" button
	document.getElementById('score').textContent= 'Average error= ' + m + '% Sample Standard Deviation= ' + sd + ' Not-moved items= ' + blank; 
	   					
	try 
  {
    if (store) {
			setStorage(m, sd, blank); 
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }    	 
  
}


// return arithmetic mean (mean, average) of array (arr) 
function mean(arr)  
{	
	let sum=0;
	for (let i in arr)  {
		sum+= parseFloat(arr[i]);
	}
	return (sum/arr.length);
}	


// return Sample Standard Deviation
// arr = array argument; mean = average (mean)
function stDev(arr, mean) 
{
	let len= arr.length, sum= 0, res;
	for (let i=0; i<len; i++) {
	  sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
	}
	res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
	return res;
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:
	 m  => The average error (%) 
   sd => Sample Standard Deviation
   not_moved => number of not-moved items  */ 
function setStorage(m, sd, not_moved)   
{    
	let oldItems = JSON.parse(localStorage.getItem('ett-v1.0')) || []; 
  let id= oldItems.length+1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(),  		   
      'error-m' : m, 	
			'error-sd' : sd,
      'not-moved': not_moved,           
			'demo' : demo  ? 1: 0,    
    	'items' : item_cnt,  			     
			'timer-time' : timer_enabled ? document.getElementById('timer').value : '',     
		 	'is-back' : is_back ? 1: 0			
  };

  oldItems.push(newItem);

  localStorage.setItem('ett-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems);   
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  let table = document.createElement("table");
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement("th");      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= "";
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement("table");  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement("th");      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('ett-v1.0')  ?  JSON.parse(localStorage.getItem('ett-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    alert('You have no ETT data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' ETT results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved ETT results:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 { 
   window.localStorage.removeItem('ett-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data  ';
 }  
}


/* when user move blue circle (dot)
   canvas => mousemove event	=> a pointing device is moved over an element (fired continuously as the mouse moves).
   this event is fired by html canvas element => document.getElementById('canvas') 
	 <canvas id="canvas" onmousemove="mouseMove(event)" onmousedown="mouseDown()" onmouseup="mouseUp()"></canvas>  */
function mouseMove(e)
{     // Note: e.offsetX == e.clientX;  e.offsetY == e.clientY 					
	let canvas= document.getElementById('canvas'), w= canvas.width, h= canvas.height;
	if ( is_move )  
	{					
		let x= e.offsetX; // mouse x (px)
		let y= e.offsetY; // mouse y (px)
		xy[item][2][0]= x; // blue circle - change current x coordinate (pixels)
		xy[item][2][1]= y; // blue circle - update current y coordinate (pixels)
		blue[item][0]= 100 * x / w; // blue circle - convert pixels into units: percents of canvas width
		blue[item][1]= 100 * y / h; // blue circle - convert pixels into units: percents of canvas height
		mis[item]= false;
		drawStim();  // repaint stimulus				
	}	
}


// html <canvas id="canvas" onmousemove="mouseMove(event)" onmousedown="mouseDown()" onmouseup="mouseUp()"></canvas>                        
function mouseDown(e)  
{
	// e.offsetX and e.offsetY => mouse pointer xy
	let x = xy[item][2][0];  // current x coordinate (pixels) for last (2.) circle 
	let y = xy[item][2][1];  // current x coordinate (pixels) for last (2.) circle			
	let d= getDistance(e.offsetX, e.offsetY, x, y); // distance from blue circle in px (current position in pixels)
	is_move= d <= 10 ? true: false;
  let canvas= document.getElementById('canvas');
	if (is_move)  {
		canvas.style.cursor= 'move';  //  'ew-resize'  or  'ns-resize'  
		canvas.style.backgroundColor= '#e5e5e5';	
	}
	else  {
	  canvas.style.backgroundColor= '#fff';	
    document.getElementById('canvas').style.cursor= 'auto'; //document.getElementById('canvas').style.cursor= 'initial';  // because of IE11		
	}	
}


// html <canvas id="canvas" onmousemove="mouseMove(event)" onmousedown="mouseDown()" onmouseup="mouseUp()"></canvas>                        
function mouseUp()  {
	is_move= false;	
	document.getElementById('canvas').style.backgroundColor= '#fff';
  document.getElementById('canvas').style.cursor= 'auto';  //document.getElementById('canvas').style.cursor= 'initial';  // because of IE11 
  drawStim();
}
 
 
/* 
Perimeter = A perimeter is either a path that encompasses/surrounds/outlines a shape (in two dimensions) or its length (one-dimensional).
Perimeter of triangle =	a + b + c 	where  a, b  and c  are the lengths of the sides of the triangle. 
Perimeter of a triangle is the sum of all its three sides. The error is shown as the deviation of the measured perimeter of the triangle from the perimeter of the optimal equilateral triangle. Unit: percentages.
The lengths of the sides of the triangle will be displayed. 
Equilateral triangle
In geometry, an equilateral triangle is a triangle in which all three sides have the same length. In the familiar Euclidean geometry, an equilateral triangle is also equiangular; that is, all three internal angles are also congruent to each other and are each 60°. It is also a regular polygon, so it is also referred to as a regular triangle. 
*/ 
